from AST import ASTNode, getParseTreeNodes
from Environment import Env
from TheTypeNode import TypeNode, TypeStruct
from CodeGenUtils import p, getLStringLabel

#   LiteralNode
#   ParamNode

# maybe add a nameNode?
#################################################################################
# literals
class LiteralNode(ASTNode):
    toLiType = dict({
        'LITERALBOOL': 'boolean',
        'LITERALCHAR': 'char',
        'LITERALSTRING': 'java.lang.String',
        'NULL': 'null',
        'NUM': 'int',
        'ZERO': 'int'
    })
    # always list all fields in the init method to show the class structure
    def __init__(self, parseTree, typeName):
        self.parseTree = parseTree
        self.name = LiteralNode.toLiType.get(parseTree.children[0].name)  # type of the literal
        self.value = parseTree.children[0].lex # the value
        self.myType = TypeStruct(self.name, None)
        self.env = None
        self.children = []
        self.typeName = typeName

        if self.name == 'int':
            self.value = int(self.value)
        if self.name == 'LITERALBOOL':
            if self.value == 'false':
                self.value = False
            else:
                self.value = True

    def linkType(self):
        if self.name == "java.lang.String":
            node = self.env.getNode(self.name, "type")
            self.myType = TypeStruct(self.name, node)

    def codeGen(self):
        if hasattr(self, "code") and self.code != "":
            return
        self.code = ""

        # boolean
        if self.name == 'boolean':
            if self.getConstant():
                self.code += p("mov", "eax", "1", " set to literal true")
                return
            else:
                self.code += p("mov", "eax", "0", " set to literal false")
                return

        # char TODO
        if self.name == 'char':
            self.code += p("mov", "eax", str(self.getConstant()), " set to literal char")
            return

        # string TODO
        # if self.name == 'java.lang.String':
        #     n = getLStringLabel()
        #     lStringLabel = "_literalstring" + n

        #     self.code += p(lStringLabel + ":", "")
        #     self.code += p("db", str(self.getConstant()))
        #     self.code += p("mov", "eax", "[" + lStringLabel + "]", " set to literal string")
        #     return

        # null
        if self.name == 'null':
            self.code += p("mov", "eax", "0", " set to literal null")
            return

        # int
        if self.name == 'int':
            self.code += p("mov", "eax", str(self.getConstant()), " set to literal int")
            return

    def getConstant(self):
        if self.name == 'int':
            return int(self.value)
        elif self.name == 'boolean':
            if self.value == 'true':
                return True
            return False
        return self.value

##################################################################################
# param type ID
class ParamNode(ASTNode):
    # always list all fields in the init method to show the class structure
    def __init__(self, parseTree, typeName):
        self.parseTree = parseTree
        self.paramType = TypeNode(parseTree.children[0], typeName)
        self.name = parseTree.children[1].lex
        self.env = None
        self.children = [self.paramType]
        self.typeName = typeName
        self.offset = 0 # offset on the stack

    def checkType(self):
        self.myType = self.paramType.myType
